// ============================================================================
// Project:	Deductive Database
// Filename:	check.h
// Purpose:	Runtime Assertion Check
// Last Change:	04.08.2017
// Language:	C++
// EMail:	brass@informatik.uni-halle.de
// WWW:		http://www.informatik.uni-halle.de/~brass/
// Address:	Feldschloesschen 15, D-06120 Halle (Saale), GERMANY
// Copyright:	(c) 1997-2017 by Stefan Brass
// License:	See file "LICENSE" for copying conditions.
// Note:	There is no warranty at all - this code may contain bugs.
// ============================================================================


/**
 * \file  check.h
 * \brief Runtime assertion check
 */


//=============================================================================
// Include File Frame:
//=============================================================================

#ifndef	CHECK_INCLUDED
#define CHECK_INCLUDED

//=============================================================================
// Used Types and Macros:
//=============================================================================

#ifndef VER_INCLUDED
#include "../base/ver.h"
#endif

#ifndef STR_INCLUDED
#include "../base/str.h"
#endif


//=============================================================================
// Exported Functions:
//=============================================================================

//-----------------------------------------------------------------------------
// check_init:
//-----------------------------------------------------------------------------

/**
 * Catch "segmentation fault", print error message to stdout.
 */

void	check_init();

//-----------------------------------------------------------------------------
// CHECK:
//-----------------------------------------------------------------------------

/**
 * Check whether an expression is true.
 */

#if	VER_DEBUG
#define	CHECK(EX, MSG) \
	(void)((EX) || (check_failed(__FILE__, __LINE__, MSG), 0))

#else
#define	CHECK(EX, MSG) \
	((void)0)
#endif

//-----------------------------------------------------------------------------
// CHECK_VALID:
//-----------------------------------------------------------------------------

/**
 * Call integrity check for current class.
 */

#if	VER_DEBUG
#define	CHECK_VALID(PLACE) \
	check_emsg(__FILE__, __LINE__, PLACE, (this==0)?"this is null":check())
#else
#define	CHECK_VALID(EX) \
	((void)0)
#endif

//-----------------------------------------------------------------------------
// CHECK_PAR:
//-----------------------------------------------------------------------------

/**
 * Check parameter, NULL is an error.
 */

#if	VER_DEBUG
#define	CHECK_PAR(PAR, PLACE) \
	check_emsg(__FILE__, __LINE__, PLACE, \
			((PAR)==0) ? "parameter is null": (PAR)->check())
#else
#define	CHECK_PAR(PAR, PLACE) \
	((void)0)
#endif

//-----------------------------------------------------------------------------
// CHECK_PAR_NULLOK:
//-----------------------------------------------------------------------------

/**
 * Check parameter, NULL is ok.
 */

#if	VER_DEBUG
#define	CHECK_PAR_NULLOK(PAR, PLACE) \
	check_emsg(__FILE__, __LINE__, PLACE, \
			((PAR)==0) ? STR_NULL: (PAR)->check())
#else
#define	CHECK_PAR_NULLOK(PAR, PLACE) \
	((void)0)
#endif

//-----------------------------------------------------------------------------
// CHECK_SUBOBJ:
//-----------------------------------------------------------------------------

/**
 * Check subobject (inside check of main object).
 */

#if	VER_DEBUG
#define	CHECK_SUBOBJ(OBJ, PLACE) \
	check_emsg(__FILE__, __LINE__, PLACE, (OBJ).check())
#else
#define	CHECK_SUBOBJ(OBJ, PLACE) \
	((void)0)
#endif

//-----------------------------------------------------------------------------
// CHECK_IMPOSSIBLE:
//-----------------------------------------------------------------------------

/**
 * Report an error.
 */

#if	VER_DEBUG
#define	CHECK_IMPOSSIBLE(MSG) \
	check_failed(__FILE__, __LINE__, MSG)
#else
#define	CHECK_IMPOSSIBLE(MSG) \
	((void)0)
#endif

//-----------------------------------------------------------------------------
// CHECK_CODE:
//-----------------------------------------------------------------------------

/**
 * Code which is only used for assertion checking.
 */

#if	VER_DEBUG
#define	CHECK_CODE(CODE) \
	CODE
#else
#define	CHECK_CODE(CODE)
#endif

//-----------------------------------------------------------------------------
// CHECK_ERR:
//-----------------------------------------------------------------------------

/**
 * An error has been detected in the check-method.
 */

#if	VER_DEBUG
#define	CHECK_ERR(MSG) \
	return (dump(), (MSG))
#else
#define	CHECK_ERR(MSG) \
	return (MSG)
#endif


//=============================================================================
// Local Functions (only used for implementation):
//=============================================================================


//-----------------------------------------------------------------------------
// check_failed:
//-----------------------------------------------------------------------------

#if	VER_DEBUG

/**
 * Print error message and abort program.
 */

void	check_failed(str_t file, int line, str_t msg);

#endif

//-----------------------------------------------------------------------------
// check_emsg:
//-----------------------------------------------------------------------------

#if	VER_DEBUG

/**
 * Check whether error message is not NULL, print and abort if yes.
 */

void	check_emsg(str_t file, int line, str_t place, str_t msg);

#endif

//-----------------------------------------------------------------------------
// check_on_exit:
//-----------------------------------------------------------------------------

#if	VER_DEBUG

/**
 *  Define a function to be called when check_failed aborts.
 */

void	check_on_exit(void (*f)());

#endif

//=============================================================================
// End of File:
//=============================================================================

#endif

