// ============================================================================
// Project:	Deductive Database
// Filename:	bam.cpp
// Purpose:	Interpreter for Bottom-Up Abstract Machine
// Last Change:	19.09.2017
// Language:	C++
// EMail:	brass@informatik.uni-halle.de
// WWW:		http://www.informatik.uni-halle.de/~brass/
// Address:	Feldschloesschen 15, D-06120 Halle (Saale), GERMANY
// Copyright:	(c) 2015-2017 by Stefan Brass
// License:	See file "LICENSE" for copying conditions.
// Note:	There is no warranty at all - this code may contain bugs.
// ============================================================================


//=============================================================================
// Interface:
//=============================================================================

#include <new>
#include <iostream>

#include "../base/ver.h"
#include "../base/str.h"
#include "../base/check.h"
#if	VER_DUMP
#include "../base/err.h"
#endif
#include "../mem/page.h"
#include "../mem/mpool.h"
#include "../bds/flexarr.h"
#include "../bds/stack.h"
#include "../dom/strmem.h"
#include "../rel/list_2.h"
#include "../rel/cur_2.h"
#include "../rel/set_2.h"
#include "../rel/set_tt.h"
#include "../rel/rel_n_n.h"
#include "../rel/cur_n_n.h"
#include "itype.h"
#include "inst.h"

#include "bam.h"


//=============================================================================
// Constructor and Destructor:
//=============================================================================


//-----------------------------------------------------------------------------
// Constructor:
//-----------------------------------------------------------------------------

bam_c::bam_c(str_t name) :
	MemPool(name),
	Stack("BAM Stack"),
	StrMem(&MemPool),
	Strings(&MemPool)
{
	// Set Magic number:
	CHECK_CODE(Magic = BAM_MAGIC);

	// Check parameter:
	CHECK(name != STR_NULL, "bam_c::constructor: name is NULL");

	// Initialize attributes (which are not nested objects):
	Error = false;
	FreeMemPtr = PAGE_U_NULL;
	FreeMemUnits = 0;
	NextCodeAddr = 0;
	NumVars = 0;
	NumRelList2 = 0;
	NumCurList2 = 0;
	NumRelSet2 = 0;
	NumRelSet2B = 0;
	NumRelMMap11 = 0;
	NumCurMMap11 = 0;
}

//-----------------------------------------------------------------------------
// Destructor:
//-----------------------------------------------------------------------------

bam_c::~bam_c()
{
	// Check this object:
	CHECK_VALID("bam_c::destructor");

	// Make this object invalid:
	CHECK_CODE(Magic = 0);
}


//=============================================================================
// Public Methods:
//=============================================================================


//-----------------------------------------------------------------------------
// vars: Set number of variables to be used.
//-----------------------------------------------------------------------------

void bam_c::vars(int num_vars) {

	// Check this object:
	CHECK_VALID("bam_c::vars");

	// Check parameter:
	CHECK(num_vars >= 0, "bam_c::vars: num_vars is negative");
	CHECK(num_vars < BAM_MAX_VARS, "bam_c::vars: num_vars too large");

	// Set number of variables:
	NumVars = num_vars;

	// Initialize variables:
	int *p = Vars;
	for(int i = num_vars; i-- > 0; )
		*p++ = -1;
}


//-----------------------------------------------------------------------------
// string: Store a string in the abstract machine string pool.
//-----------------------------------------------------------------------------

int bam_c::string(str_t str) {

	// Check this object:
	CHECK_VALID("bam_c::string");

	// Check parameter:
	CHECK(str != STR_NULL, "bam_c::string: str is NULL");

	// Find end of string:
	str_t end = str;
	while(*end)
		end++;

	// Copy the string to string memory:
	str_t copy = StrMem.insert(str, end);
	if(copy == STR_NULL) {
		Error = true;
		return -1;
	}

	// Store address of copied string in array, return index:
	int id = Strings.append(copy);
	if(id < 0)
		Error = true;
	return id;
}


//-----------------------------------------------------------------------------
// list_2:
//-----------------------------------------------------------------------------


int bam_c::list_2(list_2_t list) {

	// Check this object:
	CHECK_VALID("bam_c::list_2");

	// Check parameter:
	CHECK_PAR(list, "bam_c::list_2");

	// Is there still space?
	if(NumRelList2 >= BAM_MAX_LIST_2) {
		Error = true;
		return -1;
	}

	// Ok, store relation and return index:
	int rel_id = NumRelList2;
	RelList2[NumRelList2++] = list;
	return rel_id;
}


//-----------------------------------------------------------------------------
// set_2:
//-----------------------------------------------------------------------------


int bam_c::set_2(set_2_t set) {

	// Check this object:
	CHECK_VALID("bam_c::set_2");

	// Check parameter:
	CHECK_PAR(set, "bam_c::set_2");

	// Is there still space?
	if(NumRelSet2 >= BAM_MAX_SET_2) {
		Error = true;
		return -1;
	}

	// Ok, store relation and return index:
	int rel_id = NumRelSet2;
	RelSet2[NumRelSet2++] = set;
	return rel_id;
}


//-----------------------------------------------------------------------------
// set_2b:
//-----------------------------------------------------------------------------


int bam_c::set_2b(set_tt_t set) {

	// Check this object:
	CHECK_VALID("bam_c::set_2b");

	// Check parameter:
	CHECK_PAR(set, "bam_c::set_2b");

	// Is there still space?
	if(NumRelSet2B >= BAM_MAX_SET_2B) {
		Error = true;
		return -1;
	}

	// Ok, store relation and return index:
	int rel_id = NumRelSet2B;
	RelSet2B[NumRelSet2B++] = set;
	return rel_id;
}


//-----------------------------------------------------------------------------
// mmap_1_1: Create multimap storage structure for a relation with two columns.
//-----------------------------------------------------------------------------


int bam_c::mmap_1_1(rel_n_n_t mmap) {

	// Check this object:
	CHECK_VALID("bam_c::mmap_1_1");

	// Check parameter:
	CHECK_PAR(mmap, "bam_c::mmap_1_1");

	// Is there still space?
	if(NumRelMMap11 >= BAM_MAX_MMAP_1_1) {
		Error = true;
		return -1;
	}

	// Ok, store relation and return index:
	int rel_id = NumRelMMap11;
	RelMMap11[NumRelMMap11++] = mmap;
	return rel_id;
}


//-----------------------------------------------------------------------------
// cur_2: Create a cursor over a list with two columns.
//-----------------------------------------------------------------------------

int bam_c::cur_2(int list_id) {

	// Check this object:
	CHECK_VALID("bam_c::cur_2");

	// Check parameter:
	CHECK(list_id >= 0, "bam_c::cur_2: list_id is negative");
	CHECK(list_id < NumRelList2, "bam_c::cur_2: list_id too large");
	list_2_t list = RelList2[list_id];
	CHECK_PAR(list, "bam_c::cur_2");

	// Is there still space in the pointer array?
	if(NumCurList2 >= BAM_MAX_CUR_2) {
		Error = true;
		return -1;
	}

	// Get memory:
	if(FreeMemUnits < PAGE_UNITS(cur_2_c)) {
		page_t page = MemPool.alloc_page();
		if(page == PAGE_NULL) {
			Error = true;
			return -1;
		}
		FreeMemPtr = reinterpret_cast<page_u_t>(page);
		FreeMemUnits = PAGE_SIZE_UNITS;
	}

	// Allocate memory for cursor object:
	CHECK(FreeMemUnits >= PAGE_UNITS(cur_2_c),
		"bam_c::cur_2: Cursor object too large");
	if(FreeMemUnits < PAGE_UNITS(cur_2_c)) {
		Error = true;
		return -1;
	}
	void *mem = FreeMemPtr;
	FreeMemPtr += PAGE_UNITS(cur_2_c);
	FreeMemUnits -= PAGE_UNITS(cur_2_c);

	// Create cursor object with placement new:
	cur_2_t cursor = new (mem) cur_2_c(list);

	// Enter cursor object into array:
	int id = NumCurList2;
	CurList2[NumCurList2++] = cursor;
	return id;
}


//-----------------------------------------------------------------------------
// cur_1_1: Create a cursor over a multimap from one column to one column.
//-----------------------------------------------------------------------------

int bam_c::cur_1_1(int mmap_id) {

	// Check this object:
	CHECK_VALID("bam_c::cur_1_1");

	// Check parameter:
	CHECK(mmap_id >= 0, "bam_c::cur_1_1: mmap_id is negative");
	CHECK(mmap_id < NumRelMMap11, "bam_c::cur_1_1: mmap_id too large");
	rel_n_n_t mmap = RelMMap11[mmap_id];
	CHECK_PAR(mmap, "bam_c::cur_1_1");

	// Is there still space in the pointer array?
	if(NumCurMMap11 >= BAM_MAX_CUR_1_1) {
		Error = true;
		return -1;
	}

	// Get memory:
	if(FreeMemUnits < PAGE_UNITS(cur_n_n_c)) {
		page_t page = MemPool.alloc_page();
		if(page == PAGE_NULL) {
			Error = true;
			return -1;
		}
		FreeMemPtr = reinterpret_cast<page_u_t>(page);
		FreeMemUnits = PAGE_SIZE_UNITS;
	}

	// Allocate memory for cursor object:
	CHECK(FreeMemUnits >= PAGE_UNITS(rel_n_n_c),
		"bam_c::cur_1_1: Cursor object too large");
	if(FreeMemUnits < PAGE_UNITS(rel_n_n_c)) {
		Error = true;
		return -1;
	}
	void *mem = FreeMemPtr;
	FreeMemPtr += PAGE_UNITS(cur_n_n_c);
	FreeMemUnits -= PAGE_UNITS(cur_n_n_c);

	// Create cursor object with placement new:
	cur_n_n_t cursor = new (mem) cur_n_n_c(mmap);

	// Enter cursor object into array:
	int id = NumCurMMap11;
	CurMMap11[NumCurMMap11++] = cursor;
	return id;
}


//-----------------------------------------------------------------------------
// run: Execute a BAM program.
//-----------------------------------------------------------------------------


bool bam_c::run(int start_addr) {

	// Check this object:
	CHECK_VALID("bam_c::run");

	// Check parameter:
	CHECK(start_addr >= 0, "bam_c::run: start_addr is negative");
	CHECK(start_addr < NextCodeAddr, "bam_c::run: start_addr too large");

	// Loop over instructions:
	register unsigned char *ip = Code + start_addr;
	//err_c::dump_begin("Executed Code:");
	while(!Error) {
		//err_c::dump_int(ip-Code);
		//err_c::dump_nl();
		switch(*ip++) {
			// Null instruction (NOOP):
			case INST_NULL:
				break;

			// Instruction to stop program execution:
			case INST_HALT:
				//err_c::dump_str("HALT: ");
				//err_c::dump_str(Error?"error":"ok");
				//err_c::dump_nl();
				//dump();
				return !Error;
				break;

			// Procedure calls:
			case INST_CALL:
				{
				CHECK(ip - Code <= NextCodeAddr - 2,
					"bam_c::run: PUSH argument missing");
				int call_addr = *ip++;
				call_addr <<= 8;
				call_addr |= *ip++;
				if(!Stack.push(ip - Code)) {
					Error = true;
					return false;
				}
				ip = Code + call_addr;
				}
				break;

			case INST_RETURN:
				{
				CHECK(!Stack.is_empty(),
					"bam_c::run: POP of empty stack");
				if(Stack.is_empty()) {
					Error = true;
					return false;
				}
				int ret_addr = Stack.pop();
				CHECK(ret_addr >= 0,
					"bam_c::run: Negative return address");
				CHECK(ret_addr < NextCodeAddr,
					"bam_c::run: Return address too large");
				ip = Code + ret_addr;
				}
				break;

			// Duplicate check:
			case INST_DUPCHECK:
				CHECK_IMPOSSIBLE(
				"bam_c::run: DUPCHECK not implemented");
				break;
			case INST_DUPCHECK_2:
				{
				CHECK(ip - Code <= NextCodeAddr - 2,
				  "bam_c::run: DUPCHECK_2 arg missing");
				int set_id = *ip++;
				CHECK(set_id >= 0,
				  "bam_c::run: DUPCHECK_2 set_id<0");
				CHECK(set_id < NumRelSet2,
				  "bam_c::run: DUPCHECK_2 set_id>max");
				int var_id = *ip++;
				CHECK(var_id >= 0,
				  "bam_c::run: DUPCHECK_2 var_id<0");
				CHECK(var_id < NumVars,
				  "bam_c::run: DUPCHECK_2 var_id>max");
				CHECK(var_id + 1 < NumVars,
				  "bam_c::run: DUPCHECK_2 next var missing");
				set_2_t set = RelSet2[set_id];
				if(set->insert(Vars[var_id], Vars[var_id+1]))
					break;
				CHECK(!Stack.is_empty(),
					"bam_c::run: DUPCHECK_2: stack empty");
				if(Stack.is_empty()) {
					Error = true;
					return false;
				}
				int ret_addr = Stack.pop();
				CHECK(ret_addr >= 0,
					"bam_c::run: DUPCHECK_2 addr<0");
				CHECK(ret_addr < NextCodeAddr,
					"bam_c::run: DUPCHECK_2 addr>max");
				ip = Code + ret_addr;
				}
				break;
			case INST_DUPCHECK_2B:
				{
				CHECK(ip - Code <= NextCodeAddr - 3,
				  "bam_c::run: DUPCHECK_2B arg missing");
				int set_id = *ip++;
				CHECK(set_id >= 0,
				  "bam_c::run: DUPCHECK_2B set_id<0");
				CHECK(set_id < NumRelSet2B,
				  "bam_c::run: DUPCHECK_2B set_id>max");
				int var1_id = *ip++;
				CHECK(var1_id >= 0,
				  "bam_c::run: DUPCHECK_2B var1_id<0");
				CHECK(var1_id < NumVars,
				  "bam_c::run: DUPCHECK_2B var1_id>max");
				int var2_id = *ip++;
				CHECK(var2_id >= 0,
				  "bam_c::run: DUPCHECK_2B var2_id<0");
				CHECK(var2_id < NumVars,
				  "bam_c::run: DUPCHECK_2B var2_id>max");
				set_tt_t set = RelSet2B[set_id];
				if(set->insert(Vars[var1_id], Vars[var2_id]))
					break;
				CHECK(!Stack.is_empty(),
					"bam_c::run: DUPCHECK_2B: stack empty");
				if(Stack.is_empty()) {
					Error = true;
					return false;
				}
				int ret_addr = Stack.pop();
				CHECK(ret_addr >= 0,
					"bam_c::run: DUPCHECK_2B addr<0");
				CHECK(ret_addr < NextCodeAddr,
					"bam_c::run: DUPCHECK_2B addr>max");
				ip = Code + ret_addr;
				}
				break;

			// Saving and restoring values of variables:
			case INST_SAVE_VAR:
				{
				CHECK(ip - Code <= NextCodeAddr - 1,
				  "bam_c::run: SAVE_VAR arg missing");
				int var_id = *ip++;
				CHECK(var_id >= 0,
				  "bam_c::run: SAVE_VAR var_id<0");
				CHECK(var_id < NumVars,
				  "bam_c::run: SAVE_VAR var_id>max");
				if(!Stack.push(Vars[var_id])) {
					Error = true;
					return false;
				}
				}
				break;
			case INST_RESTORE_VAR:
				{
				CHECK(ip - Code <= NextCodeAddr - 1,
				  "bam_c::run: RESTORE_VAR arg missing");
				int var_id = *ip++;
				CHECK(var_id >= 0,
				  "bam_c::run: RESTORE_VAR var_id<0");
				CHECK(var_id < NumVars,
				  "bam_c::run: RESTORE_VAR var_id>max");
				CHECK(!Stack.is_empty(),
					"bam_c::run: RESTORE_VAR: stack empty");
				if(Stack.is_empty()) {
					Error = true;
					return false;
				}
				Vars[var_id] = Stack.pop();
				}
				break;

			// Instructions for saving and restoring cursors:
			case INST_SAVE_LIST_CUR:
				CHECK_IMPOSSIBLE(
				"bam_c::run: SAVE_LIST_CUR not implemented");
				break;
			case INST_SAVE_MMAP_CUR:
				CHECK_IMPOSSIBLE(
				"bam_c::run: SAVE_MMAP_CUR not implemented");
				break;
			case INST_SAVE_MMAP_1_1_CUR:
				{
				CHECK(ip - Code <= NextCodeAddr - 1,
				  "bam_c::run: SAVE_MMAP_1_1_CUR arg missing");
				int cur_id = *ip++;
				CHECK(cur_id >= 0,
				  "bam_c::run: SAVE_MMAP_1_1_CUR cur_id<0");
				CHECK(cur_id < NumCurMMap11,
				  "bam_c::run: SAVE_MMAP_1_1_CUR cur_id>max");
				cur_n_n_t cur = CurMMap11[cur_id];
				CHECK_PAR(cur, "bam_c::run SAVE_MMAP_1_1_CUR");
				if(!cur->push(&Stack)) {
					Error = true;
					return false;
				}
				}
				break;
			case INST_RESTORE_LIST_CUR:
				CHECK_IMPOSSIBLE(
				"bam_c::run: RESTORE_LIST_CUR not implemented");
				break;
			case INST_RESTORE_MMAP_CUR:
				CHECK_IMPOSSIBLE(
				"bam_c::run: RESTORE_MMAP_CUR not implemented");
				break;
			case INST_RESTORE_MMAP_1_1_CUR:
				{
				CHECK(ip - Code <= NextCodeAddr - 1,
				"bam_c::run: RESTORE_MMAP_1_1_CUR arg missing");
				int cur_id = *ip++;
				CHECK(cur_id >= 0,
				"bam_c::run: RESTORE_MMAP_1_1_CUR cur_id<0");
				CHECK(cur_id < NumCurMMap11,
				"bam_c::run: RESTORE_MMAP_1_1_CUR cur_id>max");
				cur_n_n_t cur = CurMMap11[cur_id];
				CHECK_PAR(cur,
					"bam_c::run RESTORE_MMAP_1_1_CUR");
				cur->pop(&Stack);
				}
				break;

			// Instructions for loops over cursors:
			case INST_LOOP_LIST:
				CHECK_IMPOSSIBLE(
				"bam_c::run: LOOP_LIST not implemented");
				break;
			case INST_LOOP_LIST_2:
				{
				CHECK(ip - Code <= NextCodeAddr - 3,
					"bam_c::run: LOOP_LIST_2 arg missing");
				int cur_id = *ip++;
				CHECK(cur_id >= 0,
					"bam_c::run: Negative cur_id");
				CHECK(cur_id < NumCurList2,
					"bam_c::run: cur_id too large");
				int not_found_addr = *ip++;
				not_found_addr <<= 8;
				not_found_addr |= *ip++;
				CHECK(not_found_addr >= 0,
					"bam_c::run: Negative end loop addr");
				CHECK(not_found_addr < NextCodeAddr,
					"bam_c::run: End loop addr too large");
				cur_2_t cur = CurList2[cur_id];
				cur->open();
				if(!cur->fetch()) {
					cur->close();
					ip = Code + not_found_addr;
				}
				}
				break;
			case INST_LOOP_MMAP:
				CHECK_IMPOSSIBLE(
				"bam_c::run: LOOP_MMAP not implemented");
				break;
			case INST_LOOP_MMAP_1_1:
				{
				CHECK(ip - Code <= NextCodeAddr - 4,
				  "bam_c::run: LOOP_MMAP_1_1 arg missing");
				int cur_id = *ip++;
				CHECK(cur_id >= 0,
				  "bam_c::run: LOOP_MMAP_1_1 cur_id<0");
				CHECK(cur_id < NumCurMMap11,
				  "bam_c::run: LOOP_MMAP_1_1 cur_id>max");
				int not_found_addr = *ip++;
				not_found_addr <<= 8;
				not_found_addr |= *ip++;
				CHECK(not_found_addr >= 0,
				  "bam_c::run: LOOP_MMAP_1_1 addr<0");
				CHECK(not_found_addr < NextCodeAddr,
				  "bam_c::run: LOOP_MMAP_1_1 addr>max");
				int var_id = *ip++;
				CHECK(var_id >= 0,
				  "bam_c::run: LOOP_MMAP_1_1 var_id<0");
				CHECK(var_id < NumVars,
				  "bam_c::run: LOOP_MMAP_1_1 var_id>max");
				cur_n_n_t cur = CurMMap11[cur_id];
				cur->open(Vars[var_id]);
				if(!cur->fetch()) {
					//cur->close();
					ip = Code + not_found_addr;
				}
				}
				break;
			case INST_GET_COL:
				CHECK_IMPOSSIBLE(
				"bam_c::run: GET_COL not implemented");
				break;
			case INST_GET_LIST_2_COL_1:
				{
				CHECK(ip - Code <= NextCodeAddr - 2,
				  "bam_c::run: GET_LIST_2_COL_1 arg missing");
				int cur_id = *ip++;
				CHECK(cur_id >= 0,
				  "bam_c::run: GET_LIST_2_COL_1 cur_id<0");
				CHECK(cur_id < NumCurList2,
				  "bam_c::run: GET_LIST_2_COL_1 cur_id>max");
				int var_id = *ip++;
				CHECK(var_id >= 0,
				  "bam_c::run: GET_LIST_2_COL_1 var_id<0");
				CHECK(var_id < NumVars,
				  "bam_c::run: GET_LIST_2_COL_1 var_id>max");
				cur_2_t cur = CurList2[cur_id];
				Vars[var_id] = cur->c1();
				}
				break;
			case INST_GET_LIST_2_COL_2:
				{
				CHECK(ip - Code <= NextCodeAddr - 2,
				  "bam_c::run: GET_LIST_2_COL_2 arg missing");
				int cur_id = *ip++;
				CHECK(cur_id >= 0,
				  "bam_c::run: GET_LIST_2_COL_2 cur_id<0");
				CHECK(cur_id < NumCurList2,
				  "bam_c::run: GET_LIST_2_COL_2 cur_id>max");
				int var_id = *ip++;
				CHECK(var_id >= 0,
				  "bam_c::run: GET_LIST_2_COL_2 var_id<0");
				CHECK(var_id < NumVars,
				  "bam_c::run: GET_LIST_2_COL_2 var_id>max");
				cur_2_t cur = CurList2[cur_id];
				Vars[var_id] = cur->c2();
				}
				break;
			case INST_GET_MMAP_1_1_OUT_1:
				{
				CHECK(ip - Code <= NextCodeAddr - 2,
				  "bam_c::run: GET_MMAP_1_1_OUT_1 arg missing");
				int cur_id = *ip++;
				CHECK(cur_id >= 0,
				  "bam_c::run: GET_MMAP_1_1_OUT_1 cur_id<0");
				CHECK(cur_id < NumCurMMap11,
				  "bam_c::run: GET_MMAP_1_1_OUT_1 cur_id>max");
				int var_id = *ip++;
				CHECK(var_id >= 0,
				  "bam_c::run: GET_MMAP_1_1_OUT_1 var_id<0");
				CHECK(var_id < NumVars,
				  "bam_c::run: GET_MMAP_1_1_OUT_1 var_id>max");
				cur_n_n_t cur = CurMMap11[cur_id];
				Vars[var_id] = cur->c2();
				}
				break;
			case INST_END_LOOP:
				CHECK_IMPOSSIBLE(
				"bam_c::run: END_LOOP not implemented");
				break;
			case INST_END_LOOP_LIST_2:
				{
				CHECK(ip - Code <= NextCodeAddr - 3,
				  "bam_c::run: END_LOOP_LIST_2 arg missing");
				int cur_id = *ip++;
				CHECK(cur_id >= 0,
					"bam_c::run: Negative cur_id");
				CHECK(cur_id < NumCurList2,
					"bam_c::run: cur_id too large");
				int loop_body_addr = *ip++;
				loop_body_addr <<= 8;
				loop_body_addr |= *ip++;
				CHECK(loop_body_addr >= 0,
					"bam_c::run: Negative loop body addr");
				CHECK(loop_body_addr < NextCodeAddr,
					"bam_c::run: Loop body addr too large");
				cur_2_t cur = CurList2[cur_id];
				if(cur->fetch())
					ip = Code + loop_body_addr;
				else
					cur->close();
				}
				break;
			case INST_END_LOOP_MMAP_1_1:
				{
				CHECK(ip - Code <= NextCodeAddr - 3,
				  "bam_c::run: END_LOOP_MMAP_1_1 arg missing");
				int cur_id = *ip++;
				CHECK(cur_id >= 0,
				  "bam_c::run: END_LOOP_MMAP_1_1 cur_id<0");
				CHECK(cur_id < NumCurMMap11,
				  "bam_c::run: END_LOOP_MMAP_1_1 cur_id>max");
				int loop_body_addr = *ip++;
				loop_body_addr <<= 8;
				loop_body_addr |= *ip++;
				CHECK(loop_body_addr >= 0,
				  "bam_c::run: END_LOOP_MMAP_1_1 addr<0");
				CHECK(loop_body_addr < NextCodeAddr,
				  "bam_c::run: END_LOOP_MMAP_1_1 addr>max");
				cur_n_n_t cur = CurMMap11[cur_id];
				if(cur->fetch())
					ip = Code + loop_body_addr;
				//else
				//	cur->close();
				}
				break;

			// Conditions:
			case INST_IF_VAR_IS:
				CHECK_IMPOSSIBLE(
				"bam_c::run: IF_VAR_IS not implemented");
				break;
			case INST_IF_VAR_EQ:
				CHECK_IMPOSSIBLE(
				"bam_c::run: IF_VAR_EQ not implemented");
				break;

			// Inserting tuples in relations:
			case INST_INSERT_LIST:
				CHECK_IMPOSSIBLE(
				"bam_c::run: INSERT_LIST not implemented");
				break;
			case INST_INSERT_LIST_2:
				{
				CHECK(ip - Code <= NextCodeAddr - 2,
				  "bam_c::run: INSERT_LIST_2 arg missing");
				int rel_id = *ip++;
				CHECK(rel_id >= 0,
				  "bam_c::run: INSERT_LIST_2 rel_id<0");
				CHECK(rel_id < NumRelList2,
				  "bam_c::run: INSERT_LIST_2 rel_id>max");
				int var1_id = *ip++;
				CHECK(var1_id >= 0,
				  "bam_c::run: INSERT_LIST_2 var1_id<0");
				CHECK(var1_id < NumVars,
				  "bam_c::run: INSERT_LIST_2 var1_id>max");
				int var2_id = *ip++;
				CHECK(var2_id >= 0,
				  "bam_c::run: INSERT_LIST_2 var2_id<0");
				CHECK(var2_id < NumVars,
				  "bam_c::run: INSERT_LIST_2 var2_id>max");
				list_2_t list = RelList2[rel_id];
				if(!list->insert(Vars[var1_id],Vars[var2_id])){
					Error = true;
					return false;
				}
				}
				break;
			case INST_INSERT_LIST_2_V:
				{
				CHECK(ip - Code <= NextCodeAddr - 2,
				  "bam_c::run: INSERT_LIST_2_V arg missing");
				int rel_id = *ip++;
				CHECK(rel_id >= 0,
				  "bam_c::run: INSERT_LIST_2_V rel_id<0");
				CHECK(rel_id < NumRelList2,
				  "bam_c::run: INSERT_LIST_2_V rel_id>max");
				int var_id = *ip++;
				CHECK(var_id >= 0,
				  "bam_c::run: INSERT_LIST_2_V var_id<0");
				CHECK(var_id < NumVars,
				  "bam_c::run: INSERT_LIST_2_V var_id>max");
				CHECK(var_id + 1 < NumVars,
				  "bam_c::run: INSERT_LIST_2_V next var missing"
				  );
				list_2_t list = RelList2[rel_id];
				if(!list->insert(Vars[var_id], Vars[var_id+1])){
					Error = true;
					return false;
				}
				}
				break;
			case INST_INSERT_MMAP:
				CHECK_IMPOSSIBLE(
				"bam_c::run: INSERT_MMAP not implemented");
				break;
			case INST_INSERT_MMAP_1_1:
				{
				CHECK(ip - Code <= NextCodeAddr - 3,
				  "bam_c::run: INSERT_MMAP_1_1 arg missing");
				int rel_id = *ip++;
				CHECK(rel_id >= 0,
				  "bam_c::run: INSERT_MMAP_1_1 rel_id<0");
				CHECK(rel_id < NumRelMMap11,
				  "bam_c::run: INSERT_MMAP_1_1 rel_id>max");
				int var1_id = *ip++;
				CHECK(var1_id >= 0,
				  "bam_c::run: INSERT_MMAP_1_1 var1_id<0");
				CHECK(var1_id < NumVars,
				  "bam_c::run: INSERT_MMAP_1_1 var1_id>max");
				int var2_id = *ip++;
				CHECK(var2_id >= 0,
				  "bam_c::run: INSERT_MMAP_1_1 var2_id<0");
				CHECK(var2_id < NumVars,
				  "bam_c::run: INSERT_MMAP_1_1 var2_id>max");
				rel_n_n_t mmap = RelMMap11[rel_id];
				if(!mmap->insert(Vars[var1_id],Vars[var2_id])){
					Error = true;
					return false;
				}
				}
				break;

			// Debugging Support:
			case INST_TRACE_CALL_2:
				{
				CHECK(ip - Code <= NextCodeAddr - 3,
				  "bam_c::run: TRACE_CALL_2 arg missing");
				int str_id = *ip++;
				CHECK(str_id >= 0,
				  "bam_c::run: TRACE_CALL_2 str_id<0");
				CHECK(str_id < Strings.length(),
				  "bam_c::run: TRACE_CALL_2 str_id>max");
				int var1_id = *ip++;
				CHECK(var1_id >= 0,
				  "bam_c::run: TRACE_CALL_2 var1_id<0");
				CHECK(var1_id < NumVars,
				  "bam_c::run: TRACE_CALL_2 var1_id>max");
				int var2_id = *ip++;
				CHECK(var2_id >= 0,
				  "bam_c::run: TRACE_CALL_2 var2_id<0");
				CHECK(var2_id < NumVars,
				  "bam_c::run: TRACE_CALL_2 var2_id>max");
				str_t pred_name = Strings.get(str_id);
				std::cout << pred_name << "(" <<
					"Var[" << var1_id << "] = " <<
					Vars[var1_id] << ", " <<
					"Var[" << var2_id << "] = " <<
					Vars[var2_id] << ")\n";
				}
				break;

			// Invalid instruction:
			default:
				CHECK_IMPOSSIBLE("run: unknown instruction");
				Error = true;
				return false;
		}

		// Check that ip is legal for next instruction:
		CHECK(ip - Code >= 0, "bam_c::run: ip too small");
		CHECK(ip - Code < NextCodeAddr, "bam_c::run: ip too large");
	}

	// This point was reached because Error was true:
	return Error;
}


//=============================================================================
// Auxiliary Methods:
//=============================================================================


//-----------------------------------------------------------------------------
// store_inst: Store an instruction without parameters.
//-----------------------------------------------------------------------------

void bam_c::store_inst(inst_t inst) {

	// Check this object:
	CHECK_VALID("bam_c::store_inst");

	// Check parameter:
	CHECK(inst_valid(inst), "bam_c::store_inst: inst is invalid");
	CHECK(inst_type(inst) == ITYPE_NONE,
			"bam_c::store_inst: wrong instruction type");

	// Check whether there is space for the instruction:
	if(NextCodeAddr >= BAM_CODE_AREA_SIZE) {
		Error = true;
		return;
	}

	// Store instruction:
	Code[NextCodeAddr++] = inst;
}


//-----------------------------------------------------------------------------
// store_inst_8: Store an instruction with an 8-bit parameter.
//-----------------------------------------------------------------------------

void bam_c::store_inst_8(inst_t inst, int arg) {

	// Check this object:
	CHECK_VALID("bam_c::store_inst_8");

	// Check parameter:
	CHECK(inst_valid(inst), "bam_c::store_inst_8: inst is invalid");
	CHECK(inst_type(inst) == ITYPE_8,
			"bam_c::store_inst_8: wrong instruction type");
	CHECK(arg >= - (0x1 << 7), "bam_c::store_inst_8: arg < -2^7");
	CHECK(arg < (0x1 << 8), "bam_c::store_inst_8: arg >= 2^8");

	// Check whether there is space for the instruction:
	if(NextCodeAddr >= BAM_CODE_AREA_SIZE - 1) {
		Error = true;
		return;
	}

	// Store instruction:
	Code[NextCodeAddr++] = inst;
	Code[NextCodeAddr++] = static_cast<unsigned char>(arg & 0xFF);
}

//-----------------------------------------------------------------------------
// store_inst_16: Store an instruction with a 16-bit parameter.
//-----------------------------------------------------------------------------

void bam_c::store_inst_16(inst_t inst, int arg) {

	// Check this object:
	CHECK_VALID("bam_c::store_inst_16");

	// Check parameter:
	CHECK(inst_valid(inst), "bam_c::store_inst_16: inst is invalid");
	CHECK(inst_type(inst) == ITYPE_16,
			"bam_c::store_inst_16: wrong instruction type");
	CHECK(arg >= - (0x1 << 15), "bam_c::store_inst_16: arg < -2^15");
	CHECK(arg < (0x1 << 16), "bam_c::store_inst_16: arg >= 2^16");

	// Check whether there is space for the instruction:
	if(NextCodeAddr >= BAM_CODE_AREA_SIZE - 2) {
		Error = true;
		return;
	}

	// Store instruction:
	Code[NextCodeAddr++] = inst;
	Code[NextCodeAddr++] = static_cast<unsigned char>((arg>>8) & 0xFF);
	Code[NextCodeAddr++] = static_cast<unsigned char>(arg & 0xFF);
}


//-----------------------------------------------------------------------------
// store_inst_8_8: Store an instruction with two 8-bit parameters.
//-----------------------------------------------------------------------------

void bam_c::store_inst_8_8(inst_t inst, int arg1, int arg2) {

	// Check this object:
	CHECK_VALID("bam_c::store_inst_8_8");

	// Check parameter:
	CHECK(inst_valid(inst), "bam_c::store_inst_8_8: inst is invalid");
	CHECK(inst_type(inst) == ITYPE_8_8,
			"bam_c::store_inst_8_8: wrong instruction type");
	CHECK(arg1 >= - (0x1 << 7), "bam_c::store_inst_8_8: arg1 < -2^7");
	CHECK(arg1 < (0x1 << 8), "bam_c::store_inst_8_8: arg1 >= 2^8");
	CHECK(arg2 >= - (0x1 << 7), "bam_c::store_inst_8_8: arg2 < -2^7");
	CHECK(arg2 < (0x1 << 8), "bam_c::store_inst_8_8: arg2 >= 2^8");

	// Check whether there is space for the instruction:
	if(NextCodeAddr >= BAM_CODE_AREA_SIZE - 2) {
		Error = true;
		return;
	}

	// Store instruction:
	Code[NextCodeAddr++] = inst;
	Code[NextCodeAddr++] = static_cast<unsigned char>(arg1 & 0xFF);
	Code[NextCodeAddr++] = static_cast<unsigned char>(arg2 & 0xFF);
}


//-----------------------------------------------------------------------------
// store_inst_8_8_8: Store an instruction with three 8-bit parameters.
//-----------------------------------------------------------------------------

void bam_c::store_inst_8_8_8(inst_t inst, int arg1, int arg2, int arg3) {

	// Check this object:
	CHECK_VALID("bam_c::store_inst_8_8_8");

	// Check parameter:
	CHECK(inst_valid(inst), "bam_c::store_inst_8_8_8: inst is invalid");
	CHECK(inst_type(inst) == ITYPE_8_8_8,
			"bam_c::store_inst_8_8_8: wrong instruction type");
	CHECK(arg1 >= - (0x1 << 7), "bam_c::store_inst_8_8_8: arg1 < -2^7");
	CHECK(arg1 < (0x1 << 8), "bam_c::store_inst_8_8_8: arg1 >= 2^8");
	CHECK(arg2 >= - (0x1 << 7), "bam_c::store_inst_8_8_8: arg2 < -2^7");
	CHECK(arg2 < (0x1 << 8), "bam_c::store_inst_8_8_8: arg2 >= 2^8");
	CHECK(arg3 >= - (0x1 << 7), "bam_c::store_inst_8_8_8: arg3 < -2^7");
	CHECK(arg3 < (0x1 << 8), "bam_c::store_inst_8_8_8: arg3 >= 2^8");

	// Check whether there is space for the instruction:
	if(NextCodeAddr >= BAM_CODE_AREA_SIZE - 3) {
		Error = true;
		return;
	}

	// Store instruction:
	Code[NextCodeAddr++] = inst;
	Code[NextCodeAddr++] = static_cast<unsigned char>(arg1 & 0xFF);
	Code[NextCodeAddr++] = static_cast<unsigned char>(arg2 & 0xFF);
	Code[NextCodeAddr++] = static_cast<unsigned char>(arg3 & 0xFF);
}

//-----------------------------------------------------------------------------
// store_inst_8_16: Store instruction with one 8-bit and one 16-bit parameter.
//-----------------------------------------------------------------------------

void bam_c::store_inst_8_16(inst_t inst, int arg1, int arg2) {

	// Check this object:
	CHECK_VALID("bam_c::store_inst_8_16");

	// Check parameter:
	CHECK(inst_valid(inst), "bam_c::store_inst_8_16: inst is invalid");
	CHECK(inst_type(inst) == ITYPE_8_16,
			"bam_c::store_inst_8_16: wrong instruction type");
	CHECK(arg1 >= - (0x1 << 7), "bam_c::store_inst_8_16: arg1 < -2^7");
	CHECK(arg1 < (0x1 << 8), "bam_c::store_inst_8_16: arg1 >= 2^8");
	CHECK(arg2 >= - (0x1 << 15), "bam_c::store_inst_8_16: arg2 < -2^7");
	CHECK(arg2 < (0x1 << 16), "bam_c::store_inst_8_16: arg2 >= 2^8");

	// Check whether there is space for the instruction:
	if(NextCodeAddr >= BAM_CODE_AREA_SIZE - 4) {
		Error = true;
		return;
	}

	// Store instruction:
	Code[NextCodeAddr++] = inst;
	Code[NextCodeAddr++] = static_cast<unsigned char>(arg1 & 0xFF);
	Code[NextCodeAddr++] = static_cast<unsigned char>((arg2>>8) & 0xFF);
	Code[NextCodeAddr++] = static_cast<unsigned char>(arg2 & 0xFF);
}


//-----------------------------------------------------------------------------
// store_inst_8_16_8: Store instruction with three parameters (8, 16, 8 bit).
//-----------------------------------------------------------------------------

void bam_c::store_inst_8_16_8(inst_t inst, int arg1, int arg2, int arg3) {

	// Check this object:
	CHECK_VALID("bam_c::store_inst_8_16_8");

	// Check parameter:
	CHECK(inst_valid(inst), "bam_c::store_inst_8_16_8: inst is invalid");
	CHECK(inst_type(inst) == ITYPE_8_16_8,
			"bam_c::store_inst_8_16_8: wrong instruction type");
	CHECK(arg1 >= - (0x1 << 7), "bam_c::store_inst_8_16_8: arg1 < -2^7");
	CHECK(arg1 < (0x1 << 8), "bam_c::store_inst_8_16_8: arg1 >= 2^8");
	CHECK(arg2 >= - (0x1 << 15), "bam_c::store_inst_8_16_8: arg2 < -2^7");
	CHECK(arg2 < (0x1 << 16), "bam_c::store_inst_8_16_8: arg2 >= 2^8");
	CHECK(arg3 >= - (0x1 << 7), "bam_c::store_inst_8_16_8: arg3 < -2^7");
	CHECK(arg3 < (0x1 << 8), "bam_c::store_inst_8_16_8: arg3 >= 2^8");

	// Check whether there is space for the instruction:
	if(NextCodeAddr >= BAM_CODE_AREA_SIZE - 5) {
		Error = true;
		return;
	}

	// Store instruction:
	Code[NextCodeAddr++] = inst;
	Code[NextCodeAddr++] = static_cast<unsigned char>(arg1 & 0xFF);
	Code[NextCodeAddr++] = static_cast<unsigned char>((arg2>>8) & 0xFF);
	Code[NextCodeAddr++] = static_cast<unsigned char>(arg2 & 0xFF);
	Code[NextCodeAddr++] = static_cast<unsigned char>(arg3 & 0xFF);
}

//-----------------------------------------------------------------------------
// store_tmp_halt: Store a temporary halt instruction.
//-----------------------------------------------------------------------------

void bam_c::store_tmp_halt() {

	// Check this object:
	CHECK_VALID("bam_c::store_tmp_halt");

	// Check whether there is space for the instruction:
	if(NextCodeAddr >= BAM_CODE_AREA_SIZE) {
		Error = true;
		return;
	}

	// Store instruction without incrementing NEXT_CODE_ADDR:
	Code[NextCodeAddr] = INST_HALT;
}


//=============================================================================
// Debugging Support:
//=============================================================================


//-----------------------------------------------------------------------------
// check: Check the integrity of this object, return error message or NULL.
//-----------------------------------------------------------------------------

#if	VER_DEBUG
str_t	bam_c::check() const
{
	// Check magic number:
	if(Magic != BAM_MAGIC)
		return "wrong magic number";

	// Check memory pool:
	CHECK_SUBOBJ(MemPool, "bam_c::check");

	// Check FreeMemUnits:
	if(FreeMemUnits < 0)
		CHECK_ERR("FreeMemUnits is negative");
	if(FreeMemUnits > PAGE_SIZE_UNITS)
		CHECK_ERR("FreeMemUnits is too large");
	if(FreeMemUnits != 0 && FreeMemPtr == PAGE_U_NULL)
		CHECK_ERR("FreeMemPtr is NULL although there should be memory");

	// Check NextCodeAddr:
	if(NextCodeAddr < 0)
		CHECK_ERR("NextCodeAddr is negative");
	if(NextCodeAddr >= BAM_CODE_AREA_SIZE)
		CHECK_ERR("NextCodeAddr is too large");

	// Check NumVars:
	if(NumVars < 0)
		CHECK_ERR("NumVars is negative");
	if(NumVars >= BAM_MAX_VARS)
		CHECK_ERR("NumVars is too large");

	// Check stack:
	CHECK_SUBOBJ(Stack, "bam_c::check");

	// Check string storage:
	CHECK_SUBOBJ(StrMem, "bam_c::check");
	CHECK_SUBOBJ(Strings, "bam_c::check");

	// Check lists with two columns:
	if(NumRelList2 < 0)
		CHECK_ERR("NumRelList2 is negative");
	if(NumRelList2 >= BAM_MAX_LIST_2)
		CHECK_ERR("NumRelList2 is too large");

	// Check cursors over lists with two columns:
	if(NumCurList2 < 0)
		CHECK_ERR("NumCurList2 is negative");
	if(NumCurList2 >= BAM_MAX_CUR_2)
		CHECK_ERR("NumCurList2 is too large");

	// Check sets with two columns:
	if(NumRelSet2 < 0)
		CHECK_ERR("NumRelSet2 is negative");
	if(NumRelSet2 >= BAM_MAX_SET_2)
		CHECK_ERR("NumRelSet2 is too large");

	// Check sets with two columns (bitmap version):
	if(NumRelSet2B < 0)
		CHECK_ERR("NumRelSet2B is negative");
	if(NumRelSet2B >= BAM_MAX_SET_2B)
		CHECK_ERR("NumRelSet2B is too large");

	// Check multimaps from one column to one column:
	if(NumRelMMap11 < 0)
		CHECK_ERR("NumRelMMap11 is negative");
	if(NumRelMMap11 >= BAM_MAX_MMAP_1_1)
		CHECK_ERR("NumRelMMap11 is too large");

	// Check cursors over multimaps from one column to one column:
	if(NumCurMMap11 < 0)
		CHECK_ERR("NumCurMMap11 is negative");
	if(NumCurMMap11 >= BAM_MAX_CUR_1_1)
		CHECK_ERR("NumCurMMap11 is too large");

	// Ok:
	return 0;
}

#endif

//-----------------------------------------------------------------------------
// dump: Show data structure.
//-----------------------------------------------------------------------------

#if	VER_DUMP

void bam_c::dump(str_t headline) const
{
	// Headline:
	if(headline == STR_NULL)
		headline = "Bottom-Up Abstract Machine (BAM)";
	err_c::dump_begin(headline);

	// Basic information:
	err_c::dump_str("Name:        '");
	err_c::dump_str(MemPool.name());
	err_c::dump_str("'");
	err_c::dump_nl();

	err_c::dump_str("Total Memory: ");
	int pages = MemPool.num_alloc_pages();
	err_c::dump_int(pages);
	err_c::dump_str(" Pages [");
	err_c::dump_int(PAGE_KB(pages));
	err_c::dump_str(" KB]");
	err_c::dump_nl();
	err_c::dump_nl();

	// Code:
	err_c::dump_str("NextCodeAddr: ");
	err_c::dump_int(NextCodeAddr);
	err_c::dump_nl();
	int addr = 0;
	while(addr < NextCodeAddr) {
		err_c::dump_str("    Code[");
		if(addr < 10)
			err_c::dump_str(" ");
		err_c::dump_int(addr);
		err_c::dump_str("]: ");
		inst_t inst = static_cast<inst_t>(Code[addr++]);
		err_c::dump_str(inst_name(inst));
		switch(inst_type(inst)) {
		case ITYPE_NONE:
			break;
		case ITYPE_8:
			{
			int arg = Code[addr++];
			err_c::dump_str("(");
			err_c::dump_int(arg);
			err_c::dump_str(")");
			}
			break;
		case ITYPE_16:
			{
			int arg = Code[addr++];
			arg <<= 8;
			arg |= Code[addr++];
			err_c::dump_str("(");
			err_c::dump_int(arg);
			err_c::dump_str(")");
			}
			break;
		case ITYPE_8_8:
			{
			int arg1 = Code[addr++];
			int arg2 = Code[addr++];
			err_c::dump_str("(");
			err_c::dump_int(arg1);
			err_c::dump_str(", ");
			err_c::dump_int(arg2);
			err_c::dump_str(")");
			}
			break;
		case ITYPE_8_8_8:
			{
			int arg1 = Code[addr++];
			int arg2 = Code[addr++];
			int arg3 = Code[addr++];
			err_c::dump_str("(");
			err_c::dump_int(arg1);
			err_c::dump_str(", ");
			err_c::dump_int(arg2);
			err_c::dump_str(", ");
			err_c::dump_int(arg3);
			err_c::dump_str(")");
			}
			break;
		case ITYPE_8_16:
			{
			int arg1 = Code[addr++];
			int arg2 = Code[addr++];
			arg2 <<= 8;
			arg2 |= Code[addr++];
			err_c::dump_str("(");
			err_c::dump_int(arg1);
			err_c::dump_str(", ");
			err_c::dump_int(arg2);
			err_c::dump_str(")");
			}
			break;
		case ITYPE_8_16_8:
			{
			int arg1 = Code[addr++];
			int arg2 = Code[addr++];
			arg2 <<= 8;
			arg2 |= Code[addr++];
			int arg3 = Code[addr++];
			err_c::dump_str("(");
			err_c::dump_int(arg1);
			err_c::dump_str(", ");
			err_c::dump_int(arg2);
			err_c::dump_str(", ");
			err_c::dump_int(arg3);
			err_c::dump_str(")");
			}
			break;
		default:
			err_c::dump_str("(*** UNKNOWN TYPE ***)");
			break;
		}
		err_c::dump_nl();
	}
	err_c::dump_nl();

	// Variables:
	err_c::dump_str("NumVars: ");
	err_c::dump_int(NumVars);
	err_c::dump_nl();
	for(int i = 0; i < NumVars; i++) {
		err_c::dump_str("    Var[");
		err_c::dump_int(i);
		err_c::dump_str("]: ");
		err_c::dump_int(Vars[i]);
		err_c::dump_nl();
	}
	err_c::dump_nl();

	// Strings:
	int num_strings = Strings.length();
	err_c::dump_str("NumStrings: ");
	err_c::dump_int(num_strings);
	err_c::dump_nl();
	for(int i = 0; i < num_strings; i++) {
		err_c::dump_str("    Str[");
		err_c::dump_int(i);
		err_c::dump_str("]: ");
		err_c::dump_str(Strings.get(i));
		err_c::dump_nl();
	}
	err_c::dump_nl();

	// Lists with two columns:
	err_c::dump_str("NumRelList2: ");
	err_c::dump_int(NumRelList2);
	err_c::dump_nl();
	for(int i = 0; i < NumRelList2; i++) {
		err_c::dump_str("    List2[");
		err_c::dump_int(i);
		err_c::dump_str("]: ");
		err_c::dump_str(RelList2[i]->name());
		err_c::dump_nl();
	}
	err_c::dump_nl();

	// Cursors for Lists with two columns:
	err_c::dump_str("NumCurList2: ");
	err_c::dump_int(NumCurList2);
	err_c::dump_nl();
	err_c::dump_nl();

	// Sets with two columns:
	err_c::dump_str("NumRelSet2: ");
	err_c::dump_int(NumRelSet2);
	err_c::dump_nl();
	for(int i = 0; i < NumRelSet2; i++) {
		err_c::dump_str("    Set2[");
		err_c::dump_int(i);
		err_c::dump_str("]: ");
		err_c::dump_str(RelSet2[i]->name());
		err_c::dump_nl();
	}
	err_c::dump_nl();

	// Sets with two columns (bitmap version):
	err_c::dump_str("NumRelSet2B: ");
	err_c::dump_int(NumRelSet2B);
	err_c::dump_nl();
	for(int i = 0; i < NumRelSet2B; i++) {
		err_c::dump_str("    Set2B[");
		err_c::dump_int(i);
		err_c::dump_str("]: ");
		err_c::dump_str(RelSet2B[i]->name());
		err_c::dump_nl();
	}
	err_c::dump_nl();

	// Multimaps with one column mapped to one column:
	err_c::dump_str("NumRelMMap11: ");
	err_c::dump_int(NumRelMMap11);
	err_c::dump_nl();
	for(int i = 0; i < NumRelMMap11; i++) {
		err_c::dump_str("    MMap11[");
		err_c::dump_int(i);
		err_c::dump_str("]: ");
		err_c::dump_str(RelMMap11[i]->name());
		err_c::dump_nl();
	}
	err_c::dump_nl();

	// Cursors for multimaps with one column mapped to one column:
	err_c::dump_str("NumCurMMap11: ");
	err_c::dump_int(NumCurMMap11);
	err_c::dump_nl();
	err_c::dump_nl();

	// Stack:
	err_c::dump_str("StackSize: ");
	err_c::dump_int(Stack.num_elems());
	err_c::dump_nl();

	// Done:
	err_c::dump_end();
}

#endif

