// ============================================================================
// Project:	Deductive Database
// Filename:	inst.cpp
// Purpose:	Enumeration Type for Abstract Machine Instructions (OpCodes)
// Last Change:	19.09.2017
// Language:	C++
// EMail:	brass@informatik.uni-halle.de
// WWW:		http://www.informatik.uni-halle.de/~brass/
// Address:	Feldschloesschen 15, D-06120 Halle (Saale), GERMANY
// Copyright:	(c) 2017 by Stefan Brass
// License:	See file "LICENSE" for copying conditions.
// Note:	There is no warranty at all - this code may contain bugs.
// ============================================================================


//=============================================================================
// Interface:
//=============================================================================

#include "../base/ver.h"
#include "../base/str.h"
#include "../base/check.h"
#include "itype.h"

#include "inst.h"

//=============================================================================
// Exported Functions:
//=============================================================================

//-----------------------------------------------------------------------------
// inst_valid: Is the parameter a valid element of the enumeration type?
//-----------------------------------------------------------------------------

#if	VER_DEBUG

bool	inst_valid(inst_t inst)
{
	return(
		// Null instruction (NOOP):
		inst == INST_NULL ||

		// Instruction to stop program execution:
		inst == INST_HALT ||

		// Procedure calls:
		inst == INST_CALL ||
		inst == INST_RETURN ||

		// Duplicate check:
		inst == INST_DUPCHECK ||
		inst == INST_DUPCHECK_2 ||
		inst == INST_DUPCHECK_2B ||

		// Saving and restoring values of variables:
		inst == INST_SAVE_VAR ||
		inst == INST_RESTORE_VAR ||

		// Instructions for saving and restoring cursors:
		inst == INST_SAVE_LIST_CUR ||
		inst == INST_SAVE_MMAP_CUR ||
		inst == INST_SAVE_MMAP_1_1_CUR ||
		inst == INST_RESTORE_LIST_CUR ||
		inst == INST_RESTORE_MMAP_CUR ||
		inst == INST_RESTORE_MMAP_1_1_CUR ||

		// Instructions for loops over cursors:
		inst == INST_LOOP_LIST ||
		inst == INST_LOOP_LIST_2 ||
		inst == INST_LOOP_MMAP ||
		inst == INST_LOOP_MMAP_1_1 ||
		inst == INST_GET_COL ||
		inst == INST_GET_LIST_2_COL_1 ||
		inst == INST_GET_LIST_2_COL_2 ||
		inst == INST_GET_MMAP_1_1_OUT_1 ||
		inst == INST_END_LOOP ||
		inst == INST_END_LOOP_LIST_2 ||
		inst == INST_END_LOOP_MMAP_1_1 ||

		// Conditions:
		inst == INST_IF_VAR_IS ||
		inst == INST_IF_VAR_EQ ||

		// Inserting tuples in relations:
		inst == INST_INSERT_LIST ||
		inst == INST_INSERT_LIST_2 ||
		inst == INST_INSERT_LIST_2_V ||
		inst == INST_INSERT_MMAP ||
		inst == INST_INSERT_MMAP_1_1 ||

		// Debugging Support:
		inst == INST_TRACE_CALL_2
	);
}

#endif


//-----------------------------------------------------------------------------
// inst_name: String representation of this instruction opcode.
//-----------------------------------------------------------------------------

str_t	inst_name(inst_t inst)
{
	str_t name;
	
	switch(inst) {
		// Null instruction (NOOP):
		case INST_NULL:
			name = "NULL";
			break;

		// Instruction to stop program execution:
		case INST_HALT:
			name = "HALT";
			break;

		// Procedure calls:
		case INST_CALL:
			name = "CALL";
			break;
		case INST_RETURN:
			name = "RETURN";
			break;

		// Duplicate check:
		case INST_DUPCHECK:
			name = "DUPCHECK";
			break;
		case INST_DUPCHECK_2:
			name = "DUPCHECK_2";
			break;
		case INST_DUPCHECK_2B:
			name = "DUPCHECK_2B";
			break;

		// Saving and restoring values of variables:
		case INST_SAVE_VAR:
			name = "SAVE_VAR";
			break;
		case INST_RESTORE_VAR:
			name = "RESTORE_VAR";
			break;

		// Instructions for saving and restoring cursors:
		case INST_SAVE_LIST_CUR:
			name = "SAVE_LIST_CUR";
			break;
		case INST_SAVE_MMAP_CUR:
			name = "SAVE_MMAP_CUR";
			break;
		case INST_SAVE_MMAP_1_1_CUR:
			name = "SAVE_MMAP_1_1_CUR";
			break;
		case INST_RESTORE_LIST_CUR:
			name = "RESTORE_LIST_CUR";
			break;
		case INST_RESTORE_MMAP_CUR:
			name = "RESTORE_MMAP_CUR";
			break;
		case INST_RESTORE_MMAP_1_1_CUR:
			name = "RESTORE_MMAP_1_1_CUR";
			break;

		// Instructions for loops over cursors:
		case INST_LOOP_LIST:
			name = "LOOP_LIST";
			break;
		case INST_LOOP_LIST_2:
			name = "LOOP_LIST_2";
			break;
		case INST_LOOP_MMAP:
			name = "LOOP_MMAP";
			break;
		case INST_LOOP_MMAP_1_1:
			name = "LOOP_MMAP_1_1";
			break;
		case INST_GET_COL:
			name = "GET_COL";
			break;
		case INST_GET_LIST_2_COL_1:
			name = "GET_LIST_2_COL_1";
			break;
		case INST_GET_LIST_2_COL_2:
			name = "GET_LIST_2_COL_2";
			break;
		case INST_GET_MMAP_1_1_OUT_1:
			name = "GET_MMAP_1_1_OUT_1";
			break;
		case INST_END_LOOP:
			name = "END_LOOP";
			break;
		case INST_END_LOOP_LIST_2:
			name = "END_LOOP_LIST_2";
			break;
		case INST_END_LOOP_MMAP_1_1:
			name = "END_LOOP_MMAP_1_1";
			break;

		// Conditions:
		case INST_IF_VAR_IS:
			name = "IF_VAR_IS";
			break;
		case INST_IF_VAR_EQ:
			name = "IF_VAR_EQ";
			break;

		// Inserting tuples in relations:
		case INST_INSERT_LIST:
			name = "INSERT_LIST";
			break;
		case INST_INSERT_LIST_2:
			name = "INSERT_LIST_2";
			break;
		case INST_INSERT_LIST_2_V:
			name = "INSERT_LIST_2_V";
			break;
		case INST_INSERT_MMAP:
			name = "INSERT_MMAP";
			break;
		case INST_INSERT_MMAP_1_1:
			name = "INSERT_MMAP_1_1";
			break;

		// Debugging Support:
		case INST_TRACE_CALL_2:
			name = "TRACE_CALL_2";
			break;

		default:
			CHECK_IMPOSSIBLE("inst_name: unknown instruction");
			name = "? (unknown/illegal)";
			break;
	}
	
	return name;
}


//-----------------------------------------------------------------------------
// inst_type: Type of the instruction based on number and length of arguments.
//-----------------------------------------------------------------------------

itype_t	inst_type(inst_t inst)
{
	itype_t itype;
	
	switch(inst) {
		// Null instruction (NOOP):
		case INST_NULL:
			itype = ITYPE_NONE;
			break;

		// Instruction to stop program execution:
		case INST_HALT:
			itype = ITYPE_NONE;
			break;

		// Procedure calls:
		case INST_CALL:
			itype = ITYPE_16;
			break;
		case INST_RETURN:
			itype = ITYPE_NONE;
			break;

		// Duplicate check:
		case INST_DUPCHECK:
			itype = ITYPE_8_8;
			break;
		case INST_DUPCHECK_2:
			itype = ITYPE_8_8;
			break;
		case INST_DUPCHECK_2B:
			itype = ITYPE_8_8_8;
			break;

		// Saving and restoring values of variables:
		case INST_SAVE_VAR:
			itype = ITYPE_8;
			break;
		case INST_RESTORE_VAR:
			itype = ITYPE_8;
			break;

		// Instructions for saving and restoring cursors:
		case INST_SAVE_LIST_CUR:
			itype = ITYPE_8;
			break;
		case INST_SAVE_MMAP_CUR:
			itype = ITYPE_8;
			break;
		case INST_SAVE_MMAP_1_1_CUR:
			itype = ITYPE_8;
			break;
		case INST_RESTORE_LIST_CUR:
			itype = ITYPE_8;
			break;
		case INST_RESTORE_MMAP_CUR:
			itype = ITYPE_8;
			break;
		case INST_RESTORE_MMAP_1_1_CUR:
			itype = ITYPE_8;
			break;

		// Instructions for loops over cursors:
		case INST_LOOP_LIST:
			itype = ITYPE_8_16;
			break;
		case INST_LOOP_LIST_2:
			itype = ITYPE_8_16;
			break;
		case INST_LOOP_MMAP:
			itype = ITYPE_8_16;
			break;
		case INST_LOOP_MMAP_1_1:
			itype = ITYPE_8_16_8;
			break;
		case INST_GET_COL:
			itype = ITYPE_8_8;
			break;
		case INST_GET_LIST_2_COL_1:
			itype = ITYPE_8_8;
			break;
		case INST_GET_LIST_2_COL_2:
			itype = ITYPE_8_8;
			break;
		case INST_GET_MMAP_1_1_OUT_1:
			itype = ITYPE_8_8;
			break;
		case INST_END_LOOP:
			itype = ITYPE_8_16;
			break;
		case INST_END_LOOP_LIST_2:
			itype = ITYPE_8_16;
			break;
		case INST_END_LOOP_MMAP_1_1:
			itype = ITYPE_8_16;
			break;

		// Conditions:
		case INST_IF_VAR_IS:
			itype = ITYPE_8_16;
			break;
		case INST_IF_VAR_EQ:
			itype = ITYPE_8_8;
			break;

		// Inserting tuples in relations:
		case INST_INSERT_LIST:
			itype = ITYPE_8_8;
			break;
		case INST_INSERT_LIST_2:
			itype = ITYPE_8_8_8;
			break;
		case INST_INSERT_LIST_2_V:
			itype = ITYPE_8_8;
			break;
		case INST_INSERT_MMAP:
			itype = ITYPE_8_8;
			break;
		case INST_INSERT_MMAP_1_1:
			itype = ITYPE_8_8_8;
			break;

		// Debugging Support:
		case INST_TRACE_CALL_2:
			itype = ITYPE_8_8_8;
			break;

		default:
			CHECK_IMPOSSIBLE("inst_name: unknown instruction");
			itype = ITYPE_NONE;
			break;
	}
	
	return itype;
}

