// ============================================================================
// Project:	Deductive Database
// Filename:	gen_cur2.cpp
// Purpose:	Cursor over generated two-column relations (for tests).
// Last Change:	04.08.2017
// Language:	C++
// EMail:	brass@informatik.uni-halle.de
// WWW:		http://www.informatik.uni-halle.de/~brass/
// Address:	Feldschloesschen 15, D-06120 Halle (Saale), GERMANY
// Copyright:	(c) 2016-2017 by Stefan Brass
// License:	See file "LICENSE" for copying conditions.
// Note:	There is no warranty at all - this code may contain bugs.
// ============================================================================


//=============================================================================
// Interface:
//=============================================================================

#include "../base/ver.h"
#include "../base/str.h"
#if	VER_DUMP
#include "../base/err.h"
#endif
#include "../base/check.h"

#include "gen_cur2.h"


//=============================================================================
// Constructor and Destructor:
//=============================================================================


//-----------------------------------------------------------------------------
// Constructor 1 with full parameter selection:
//-----------------------------------------------------------------------------

gen_cur2_c::gen_cur2_c(int size, int c1_start, int c1_step, int c1_max,
			int c2_start, int c2_step, int c2_max)
{
	// Check parameters:
	CHECK(size > 0,
		"gen_cur2_c::constructor: size must be at least 1");
	CHECK(c1_step >= 1,
		"gen_cur2_c::constructor: c1_step must be at least 1");
	CHECK(c2_step >= 1,
		"gen_cur2_c::constructor: c2_step must be at least 1");
	CHECK(c1_start <= c1_max,
		"gen_cur2_c::constructor: c1_max must be >= c1_start");
	CHECK(c2_start <= c2_max,
		"gen_cur2_c::constructor: c2_max must be >= c2_start");
	CHECK(size <= c1_max - c1_start + 1,
		"gen_cur2_c::constructor: c1_max too small or size too big");
	CHECK(size <= c2_max - c2_start + 1,
		"gen_cur2_c::constructor: c2_max too small or size too big");

	// Set Magic number:
	CHECK_CODE(Magic = GEN_CUR2_MAGIC);

	// Initialize attributes that are never changed during object lifetime:
	Size   = size;
	Start1 = c1_start;
	Step1  = c1_step;
	Max1   = c1_max;
	Start2 = c2_start;
	Step2  = c2_step;
	Max2   = c2_max;

	// Initialize sequence generator attributes and current tuple:
	init();
}

//-----------------------------------------------------------------------------
// Constructor 2 with simplified relation ID:
//-----------------------------------------------------------------------------

gen_cur2_c::gen_cur2_c(int rel_id)
{
	// Check parameters:
	CHECK(rel_id >= 0 && rel_id <= 4,
		"gen_cur2_c::constructor2: there are only relations 1-4");

	// Set Magic number:
	CHECK_CODE(Magic = GEN_CUR2_MAGIC);

	// Set relation parameters:
	switch(rel_id) {
		case 0:
			// Simple relation with 10 tuples:
			Size   = 10;
			Start1 = 1;
			Step1  = 1;
			Max1   = 10;
			Start2 = 100;
			Step2  = 5;
			Max2   = 200;
			break;
		case 1:
			// Simple relation with one million tuples:
			Size   = 1000000;
			Start1 = 1;
			Step1  = 1;
			Max1   = 1000000;
			Start2 = 1;
			Step2  = 1;
			Max2   = 1000000;
			break;
		case 2:
			// Relation with more "jumps", one million rows:
			Size   = 1000000;
			Start1 = 1;
			Step1  = 137;
			Max1   = 1000000;
			Start2 = 1000;
			Step2  = 1001;
			Max2   = 2000000;
			break;
		case 3:
			// Relation with more "jumps", 100 million rows:
			Size   = 100000000;
			Start1 = 1;
			Step1  = 137;
			Max1   = 100000000;
			Start2 = 1000;
			Step2  = 1001;
			Max2   = 200000000;
			break;
		case 4:
			// Relation with more "jumps", one billion rows:
			Size   = 1000000000;
			Start1 = 1;
			Step1  = 137;
			Max1   = 1000000000;
			Start2 = 1000;
			Step2  = 1001;
			Max2   = 2000000000;
			break;
		default:
			CHECK_IMPOSSIBLE(
				"gen_cur2_c::constructor2: invalid rel_id");
			Size   = 1;
			Start1 = 1;
			Step1  = 1;
			Max1   = 1;
			Start2 = 1;
			Step2  = 1;
			Max2   = 1;
			break;
	}

	// Initialize sequence generator attributes and current tuple:
	init();
}

//-----------------------------------------------------------------------------
// Destructor:
//-----------------------------------------------------------------------------

gen_cur2_c::~gen_cur2_c()
{
	// Check this object:
	CHECK_VALID("gen_cur2_c::destructor");

	// Make this object invalid:
	CHECK_CODE(Magic = 0);
}


//-----------------------------------------------------------------------------
// open: Open the cursor.
//-----------------------------------------------------------------------------

void gen_cur2_c::open()
{
	// Check this object:
	CHECK_VALID("gen_cur2_c::open");

	// Initialize sequence generator attributes and current tuple:
	init();

	// The tuple number must be 0 (cursor open, just before first tuple):
	TupleNo = 0;
}



//=============================================================================
// Auxiliary Methods:
//=============================================================================

//-----------------------------------------------------------------------------
// init: Initialize sequence generator attributes and current tuple.
//-----------------------------------------------------------------------------

void gen_cur2_c::init()
{
	// Initialize attributes to represent current tuple (cursor closed):
	TupleNo = -1;
	C1      = -1;
	C2      = -1;

	// Initialize iteration counter:
	Iter1   = 1;
	Iter2   = 1;
}

//=============================================================================
// Debugging Support:
//=============================================================================

//-----------------------------------------------------------------------------
// check: Check the integrity of this object, return error message or NULL.
//-----------------------------------------------------------------------------

#if	VER_DEBUG
str_t	gen_cur2_c::check() const
{
	// Check magic number:
	if(Magic != GEN_CUR2_MAGIC)
		return "wrong magic number";

	// Check parameters of sequence:
	if(Size <= 0)
		return "Size must be at least 1";
	if(Step1 < 1)
		return "Step1 must be at least 1";
	if(Step2 < 1)
		return "Step2 must be at least 1";
	if(Start1 > Max1)
		return "Max1 must be >= Start1";
	if(Start2 > Max2)
		return "Max2 must be >= Start2";
	if(Size > Max1 - Start1 + 1)
		return "Max1 too small or size too big";
	if(Size > Max2 - Start2 + 1)
		return "Max2 too small or size too big";

	// Check tuple number and data values:
	if(TupleNo < -1) {
		return "TupleNo is negative and not -1";
	}
	else if(TupleNo == -1 || TupleNo == 0) {
		if(C1 != -1)
			return "cursor closed, but C1 != -1";
		if(C2 != -1)
			return "cursor closed, but C2 != -1";
	}
	else {
		if(C1 < Start1)
			return "cursor open, but C1 is below Start1";
		if(C1 > Max1)
			return "C1 is too large";
		if(C2 < Start2)
			return "cursor open, but C2 is below Start2";
		if(C2 > Max2)
			return "C2 is too large";
		if(TupleNo > Size + 1)
			return "cursor has run behind end of relation";
	}

	// Ok:
	return 0;
}
#endif


//-----------------------------------------------------------------------------
// dump: Show data structure.
//-----------------------------------------------------------------------------

#if	VER_DUMP
void	gen_cur2_c::dump(str_t headline) const
{
	// Headline:
	if(headline == STR_NULL)
		headline =
		"Curor for test data (Relation with two integer columns)";
	err_c::dump_begin(headline);

	// Relation Parameters:
	err_c::dump_str("Relation Parameters:");
	err_c::dump_nl();
	err_c::dump_str("    Size:    ");
	err_c::dump_int(Size);
	err_c::dump_nl();
	err_c::dump_str("    Start1:  ");
	err_c::dump_int(Start1);
	err_c::dump_nl();
	err_c::dump_str("    Step1:   ");
	err_c::dump_int(Step1);
	err_c::dump_nl();
	err_c::dump_str("    Max1:    ");
	err_c::dump_int(Max1);
	err_c::dump_nl();
	err_c::dump_str("    Start2:  ");
	err_c::dump_int(Start2);
	err_c::dump_nl();
	err_c::dump_str("    Step2:   ");
	err_c::dump_int(Step2);
	err_c::dump_nl();
	err_c::dump_str("    Max2:    ");
	err_c::dump_int(Max2);
	err_c::dump_nl();
	err_c::dump_nl();

	// Current Tuple:
	err_c::dump_str("Current Tuple:");
	err_c::dump_nl();
	err_c::dump_str("    TupleNo: ");
	err_c::dump_int(TupleNo);
	err_c::dump_nl();
	err_c::dump_str("    C1:      ");
	err_c::dump_int(C1);
	err_c::dump_nl();
	err_c::dump_str("    C2:      ");
	err_c::dump_int(C2);
	err_c::dump_nl();
	err_c::dump_nl();

	// Sequence Generator:
	err_c::dump_str("Sequence Generator:");
	err_c::dump_nl();
	err_c::dump_str("    Iter1:   ");
	err_c::dump_int(Iter1);
	err_c::dump_nl();
	err_c::dump_str("    Iter2:   ");
	err_c::dump_int(Iter2);
	err_c::dump_nl();
	err_c::dump_nl();

	// Size of this object:
	err_c::dump_str("Object Size:    ");
	err_c::dump_uint(static_cast<unsigned int>(sizeof(gen_cur2_c)));
	err_c::dump_str(" Bytes");
	err_c::dump_nl();
	err_c::dump_nl();

	// Done:
	err_c::dump_end();
}
#endif

