// ============================================================================
// Project:	Deductive Database
// Filename:	load_tc.cpp
// Purpose:	Parse and load data for "parent" relation for tc benchmark.
// Last Change:	04.08.2017
// Language:	C++
// EMail:	brass@informatik.uni-halle.de
// WWW:		http://www.informatik.uni-halle.de/~brass/
// Address:	Feldschloesschen 15, D-06120 Halle (Saale), GERMANY
// Copyright:	(c) 2015-2017 by Stefan Brass
// License:	See file "LICENSE" for copying conditions.
// Note:	There is no warranty at all - this code may contain bugs.
// ============================================================================


//=============================================================================
// This Module is optional:
//=============================================================================

#include "../base/ver.h"
#if	VER_TCFF_LOAD


//=============================================================================
// Interface:
//=============================================================================

#include "../base/str.h"
#include "../base/err.h"
#include "../base/check.h"
#include "../input/in_file.h"
#include "../facts/syn.h"
#include "../facts/lex.h"
#include "../rel/list_2.h"
#include "../rel/rel_n_n.h"
#include "out.h"

#include "load_tc.h"


//=============================================================================
// Static Methods:
//=============================================================================


//-----------------------------------------------------------------------------
// Load data:
//-----------------------------------------------------------------------------

bool load_tc_c::load(
		// Name of data file to load:
		str_t filename,
		// Relations for EDB-literals:
		list_2_t  par_ff,
		rel_n_n_t par_fb)
{
	register str_t p;

	// Open file:
	in_file_c file(filename);
	if(!file.is_open()) {
		// Error message was printed in constructor.
		return false;
	}

	// Create Scanner Object:
	lex_c scanner(&file);
		// This reads already first token.

	// Variables for fact arguments:
	int arg1;
	int arg2;

	// Main loop (over Datalog facts in file):
	while(scanner.tok_type() != TOK_EOF) {
		// Predicate:
		if(scanner.tok_type() != TOK_ATOM) {
			scanner.show_err(SYN_PRED_EXPECTED);
			return false;
		}
		p = scanner.tok_ptr();

		// There is only one predicate in the data file: 'par'
		if((scanner.tok_end() - p) != 3 ||
			((p[0] != 'p') | (p[1] != 'a') | (p[2] != 'r'))) {
			scanner.show_err(SYN_UNKNOWN_PRED);
			return false;
		}

		// Left parenthesis '(':
		scanner.next();
		if(scanner.tok_type() != TOK_LPAREN) {
			scanner.show_err(SYN_LPAREN_EXPECTED);
			return false;
		}

		// First argument (must be an integer):
		scanner.next();
		if(scanner.tok_type() != TOK_INT) {
			scanner.show_err(SYN_ARG_MUST_BE_INT);
			return false;
		}
		arg1 = scanner.tok_int();
		if(arg1 < 0) {
			scanner.show_err(SYN_ARG_CANNOT_BE_NEGATIVE);
			return false;
		}

		// Comma to separate arguments:
		scanner.next();
		if(scanner.tok_type() != TOK_COMMA) {
			scanner.show_err(SYN_COMMA_EXPECTED);
			return false;
		}

		// Second argument (must be an integer):
		scanner.next();
		if(scanner.tok_type() != TOK_INT) {
			scanner.show_err(SYN_ARG_MUST_BE_INT);
			return false;
		}
		arg2 = scanner.tok_int();
		if(arg2 < 0) {
			scanner.show_err(SYN_ARG_CANNOT_BE_NEGATIVE);
			return false;
		}

		// Right parenthesis ')':
		scanner.next();
		if(scanner.tok_type() != TOK_RPAREN) {
			scanner.show_err(SYN_RPAREN_EXPECTED);
			return false;
		}

		// Full Stop '.':
		scanner.next();
		if(scanner.tok_type() != TOK_FULLSTOP) {
			scanner.show_err(SYN_FULLSTOP_EXPECTED);
			return false;
		}

		// Insert tuple (all tuples are relevant):
		par_ff->insert(arg1, arg2);
		par_fb->insert(arg2, arg1);

		// Read next token:
		scanner.next();

	} // End of Loop Over Facts

	// Read the entire file without error:
	return true;
}


//=============================================================================
// End of Conditional Compilation of this Test (VER_TCFF_LOAD):
//=============================================================================

#endif

